open Printf
open Ast
open Typechecker

let headerlib = "lib/header.mrlib"
let main_template = "lib/main_template.mrlib"
let mapper_template = "lib/mapper_template.mrlib"
let reducer_template = "lib/reducer_template.mrlib"


module StringMap = Map.Make(String)

let quote s =
    "\"" ^ s ^ "\""

let readlines filename = 
	let lines = ref [] in
	let channel = open_in filename in
	try
  		while true; do
    	lines := input_line channel :: !lines
  	done; []
	with End_of_file ->	close_in channel; List.rev !lines

let copy_header (to_channel) = 
	let lines = readlines headerlib in
	List.iter (fun line -> (output_string to_channel (line ^ "\n"))) lines

let get_job_name = function
	Config_dec(attribute, job_name) -> job_name

let get_type_name = function
		Atom_type_spc(Int) -> "IntWritable "
	|	Atom_type_spc(Double) -> "DoubleWritable "
	|	Atom_type_spc(Bool) -> "BooleanWritable "
	|	Atom_type_spc(Text) -> "Text "
	|	_ -> raise (Failure ("unexpected list type as input/output key/value"))

let get_internal_type_name = function
		Atom_type_spc(Int) -> "int "
	|   Atom_type_spc(Double) -> "double "
	|   Atom_type_spc(Bool) -> "boolean "
	|   Atom_type_spc(Text) -> "String "
	|	List_type_spc(Int) -> "int[] "
	|	List_type_spc(Double) -> "double[] "
	|	List_type_spc(Bool) -> "bool[] "
	|	List_type_spc(Text) -> "String[] "

let get_input_key = function
	Map_rel(input_key, input_value, output_key, output_value) -> input_key

let get_input_value = function
	Map_rel(input_key, input_value, output_key, output_value) -> input_value

let get_output_key = function
	Map_rel(input_key, input_value, output_key, output_value) -> output_key

let get_output_value = function
	Map_rel(input_key, input_value, output_key, output_value) -> output_value

let get_mapper_name configuration =
	(get_job_name configuration) ^ "Mapper"

let get_reducer_name configuration =
	(get_job_name configuration) ^ "Reducer"

let get_file_name (program) =
	let get_configuration = function
		Program(configuration, mapper, reducer) -> configuration 
	in (get_job_name (get_configuration program)) ^ ".java"

let translate_op = function
		Add		-> 	"+"
	|	Sub		->	"-"
	|	Mul		->	"*"
	|	Div		->	"/"
	|	Lt		->	"<"
	|	Leq		->	"<="
	|	Gt		->	">"
	|	Geq		->	">="
	|	Equal	->	"=="
	|	Neq		->	"!="
	|	And		->	"&&"
	|	Or		->	"||"

let add_symbol mr_name java_name mr_type symbol_table =
	symbol_table := StringMap.add mr_name (java_name, mr_type) !symbol_table; !symbol_table

let remove_symbol mr_name symbol_table =
	symbol_table := StringMap.remove mr_name !symbol_table; !symbol_table

let get_symbol mr_name symbol_table =
	try
		StringMap.find mr_name !symbol_table
	with Not_found -> raise (Failure ("undeclared variable " ^ mr_name))

let get_java_name mr_name symbol_table =
	let (java_name, mr_type) = get_symbol mr_name symbol_table in java_name

let get_java_type mr_name symbol_table =
	let (java_name, mr_type) = get_symbol mr_name symbol_table in get_internal_type_name mr_type

let get_element_type id symbol_table = 
    let (_, mr_type) = get_symbol id symbol_table in
	match mr_type with
        List_type_spc(Int) -> Atom_type_spc(Int)
    |   List_type_spc(Double) -> Atom_type_spc(Double)
    |   List_type_spc(Bool) -> Atom_type_spc(Bool)
    |   List_type_spc(Text) -> Atom_type_spc(Text)
    |   _ -> raise (Failure ( id ^ " is not a List"))

let get_list_type = function
		Atom_type_spc(Int) -> List_type_spc(Int)
	|	Atom_type_spc(Double) -> List_type_spc(Double)
	|	Atom_type_spc(Bool)	-> List_type_spc(Bool)
	|	Atom_type_spc(Text)	-> List_type_spc(Text)
	|	_ -> raise (Failure ("Unexpected List Type"))

let get_type_of_name id symbol_table =
	let (_, mr_type) = get_symbol id symbol_table in
	get_type_name mr_type

let rec translate_expression symbol_table = function
		Id(id)					-> get_java_name id symbol_table
	|	Integer_Const(v)		-> (string_of_int v)
	|	Double_Const(v)			-> (string_of_float v)
	|	Text_Const(v)			-> "\"" ^ v ^ "\""
	|	Bool_Const(v)			-> (string_of_bool v)
	|	Uminus(expr)			-> "-" ^ translate_expression symbol_table expr
	|	Split(id, delimitter) 	-> (get_java_name id symbol_table) ^ "." ^ "split(" ^ (quote delimitter) ^ ")"
	|	Assign(id, expr)		-> (get_java_name id symbol_table) ^ "=" ^ translate_expression symbol_table expr
	|	Vdcl(type_spc, id)		-> ignore(add_symbol id id type_spc symbol_table); ((get_internal_type_name type_spc) ^ id)
	|	Binop(expr1, op, expr2)	-> ((translate_expression symbol_table expr1) ^ (translate_op op) ^ (translate_expression symbol_table expr2))
	|	Vdfn(type_spc, id, expr)-> ignore(add_symbol id id type_spc symbol_table); ((get_internal_type_name type_spc) ^ id) ^ " = " ^ (translate_expression symbol_table expr)

let rec translate_statement symbol_table = function
		Block(statements) 		-> "{\n" ^ (translate_statements symbol_table statements) ^ "}\n"
	|	Expr(expr)				-> (translate_expression symbol_table expr) ^ ";\n"
	|	Emit(id1, id2)			-> "context.write(new " ^ (get_type_of_name id1 symbol_table) ^ "(" ^ (get_java_name id1 symbol_table) ^ "), new " ^ (get_type_of_name id2 symbol_table) ^ "(" ^ (get_java_name id2 symbol_table) ^ "));\n"
	|	If(expr, stmt1, stmt2)	-> "if(" ^ (translate_expression symbol_table expr) ^ ")" ^ (translate_statement symbol_table stmt1) ^ "else\n" ^ (translate_statement symbol_table stmt2)
	|	Foreach(id1, id2, stmt)	-> 
		let element_type = get_element_type id2 symbol_table in
		let list_name = get_java_name id2 symbol_table in
		let _ = add_symbol id1 id1 element_type symbol_table in
		let statement = "for(" ^ (get_internal_type_name element_type) ^ id1 ^ " : " ^ list_name ^ ")\n" ^ (translate_statement symbol_table stmt) ^ "\n" in
		let _ = remove_symbol id1 symbol_table in
		statement

and translate_statement_aux symbol_table accumulator statement =
	let translator = translate_statement symbol_table in
	accumulator ^ (translator statement)

and translate_statements symbol_table statements =
	let translator = translate_statement_aux symbol_table in
	List.fold_left translator "" statements

external string_to_format:
	string -> ('a, 'b, 'c, 'd, 'e, 'f) format6 = "%identity"

let translate_para_as_atom para = function
		Atom_type_spc(Int) 		-> "int " ^ para ^ "__" ^ "=" ^ para ^ ".get();\n"
	|   Atom_type_spc(Double) 	-> "double " ^ para ^ "__" ^ "=" ^ para ^ ".get();\n"
 	|   Atom_type_spc(Bool) 	-> "boolean " ^ para ^ "__" ^ "=" ^ para ^ ".get();\n"
 	|   Atom_type_spc(Text) 	-> "String " ^ para ^ "__" ^ "=" ^ para ^ ".toString();\n"
 	|   _ -> raise (Failure ("unexpected list type as input/output key/value"))

let copy_para para mr_type =
	"for (" ^ (get_type_name mr_type) ^ "__tmp__ : " ^ para ^ ")\n" ^ para ^ "__.add(__tmp__.get());\n"

let translate_para_as_list para mr_type =
	match mr_type with
		Atom_type_spc(Int)		-> 	"ArrayList<Integer> " ^ para ^ "__ = " ^ "new ArrayList<Integer>();\n" ^ copy_para para mr_type
	|	Atom_type_spc(Double)	->	"ArrayList<Double> " ^ para ^ "__ = " ^ "new ArrayList<Double>();\n" ^ copy_para para mr_type
	|	Atom_type_spc(Bool)		->	"ArrayList<Boolean> " ^ para ^ "__ = " ^ "new ArrayList<Boolean>();\n" ^ copy_para para mr_type
	|	Atom_type_spc(Text)		->	"ArrayList<String> " ^ para ^ "__ = " ^ "new ArrayList<String>();\n" ^ copy_para para mr_type
	|	_ -> raise (Failure ("unable to convert " ^ para ^" as a list in Java type"))
	

let translate_mapper configuration (mapper:Ast.mapper_def) (to_channel:out_channel) = 
	let symbol_table = ref StringMap.empty in
	let class_name = get_mapper_name configuration in
	let input_value_type = get_input_value mapper.map_rel in
	let output_key_type_name = get_type_name (get_output_key mapper.map_rel) in
	let output_value_type_name = get_type_name (get_output_value mapper.map_rel) in
	let para1 = List.nth mapper.map_paras 0 in
	let para2 = List.nth mapper.map_paras 1 in
	let translated_para2 = translate_para_as_atom para2 input_value_type in
	let _ = add_symbol para2 (para2 ^ "__") input_value_type symbol_table in
	let function_body = translate_statements symbol_table mapper.map_body in
	let lines = readlines mapper_template in
	let format_string = String.concat "\n" lines in
	fprintf to_channel (string_to_format format_string)
						class_name
						output_key_type_name
						output_value_type_name
						para1
						para2
						(translated_para2 ^ function_body)

let translate_reducer configuration (reducer:Ast.reducer_def) (to_channel:out_channel) = 
	let symbol_table = ref StringMap.empty in
	let class_name = get_reducer_name configuration in
	let input_key_type = get_input_key reducer.red_rel
	and input_value_type = get_input_value reducer.red_rel
	and output_key_type = get_output_key reducer.red_rel
	and output_value_type = get_output_value reducer.red_rel in
	let input_key_type_name = get_type_name input_key_type
	and input_value_type_name = get_type_name input_value_type
	and output_key_type_name = get_type_name output_key_type
	and output_value_type_name = get_type_name output_value_type in
	let para1 = List.nth reducer.red_paras 0 in
	let para2 = List.nth reducer.red_paras 1 in
	let translated_para1 = translate_para_as_atom para1 input_key_type in
	let translated_para2 = translate_para_as_list para2 input_value_type in
	let _ = add_symbol para1 (para1 ^ "__") input_key_type symbol_table in
	let _ = add_symbol para2 (para2 ^ "__") (get_list_type input_value_type) symbol_table in
	let function_body = translate_statements symbol_table reducer.red_body in
	let lines = readlines reducer_template in
	let format_string = String.concat "\n" lines in
	fprintf to_channel (string_to_format format_string)
						class_name
						input_key_type_name
						input_value_type_name
						output_key_type_name
						output_value_type_name
						input_key_type_name
						para1
						input_value_type_name
						para2
						(translated_para1 ^ translated_para2 ^ function_body)

let translate_main configuration mapper reducer to_channel =
	let job_name = get_job_name configuration in
	let lines = readlines main_template in
	let format_string = String.concat "\n" lines in
	fprintf to_channel (string_to_format format_string)
						(quote job_name) 					(* job name *)
						job_name  							(* main class *)
						(get_mapper_name configuration) 	(* mapper class *)
						(get_reducer_name configuration)	(* reducer class *)
						(get_type_name (get_output_key reducer.red_rel))	(* output key *)
						(get_type_name (get_output_value reducer.red_rel))	(* output value *)

let wrap_main_class_open configuration to_channel =
	output_string to_channel ("public class " ^ (get_job_name configuration)  ^ "{\n")

let wrap_main_class_close configuration to_channel =
	output_string to_channel "}\n"

let rec translate (to_channel:out_channel) = function
    Program(configuration, mapper, reducer) -> 
			copy_header to_channel;
			wrap_main_class_open configuration to_channel; 
			ignore(translate_mapper configuration mapper to_channel);
			ignore(translate_reducer configuration reducer to_channel);
			ignore(translate_main configuration mapper reducer to_channel);
			wrap_main_class_close configuration to_channel
(* main *)
let _ =
	let lexbuf = Lexing.from_channel stdin in
	let program = Parser.program Scanner.token lexbuf in
	let file = open_out (get_file_name program) in 
	(*let _ = Typechecker.typechecker program in*)
	let _ = translate file program in
	close_out file
